@@ -5,7 +5,8 @@ from __future__ import absolute_import
 from six.moves import range
 from six.moves import zip
 
-__copyright__ = """Copyright 2011-2012 Andreas Kloeckner"""
+__copyright__ = """Copyright 2011-2012 Andreas Kloeckner \
+                   Copyright 2017 Hao Gao"""
 
 __license__ = """
 Permission is hereby granted, free of charge, to any person
@@ -300,6 +301,7 @@ def _make_sort_scan_type(device, bits, index_dtype):
     dtype = get_or_register_dtype(name, dtype)
     return name, dtype, c_decl
 
+
 # {{{ types, helpers preamble
 
 RADIX_SORT_PREAMBLE_TPL = Template(r"""//CL//
@@ -523,7 +525,7 @@ class RadixSort(object):
         if allocator is None:
             allocator = args[self.first_array_arg_idx].allocator
 
-        queue = kwargs.pop("allocator", None)
+        queue = kwargs.pop("queue", None)
         if queue is None:
             queue = args[self.first_array_arg_idx].queue
 
@@ -645,8 +647,18 @@ void ${kernel_name}(${kernel_list_arg_decl} USER_ARG_DECL index_type n)
         %else:
             %for name, dtype in list_names_and_dtypes:
                 %if name not in count_sharing:
-                    index_type plb_${name}_index =
-                        plb_${name}_start_index[i];
+                    index_type plb_${name}_index;
+                    if (plb_${name}_start_index)
+                        %if eliminate_empty_output_lists:
+                            plb_${name}_index =
+                                plb_${name}_start_index[
+                                    ${name}_compressed_indices[i]
+                                ];
+                        %else:
+                            plb_${name}_index = plb_${name}_start_index[i];
+                        %endif
+                    else
+                        plb_${name}_index = 0;
                 %endif
             %endfor
         %endif
@@ -656,7 +668,8 @@ void ${kernel_name}(${kernel_list_arg_decl} USER_ARG_DECL index_type n)
         %if is_count_stage:
             %for name, dtype in list_names_and_dtypes:
                 %if name not in count_sharing:
-                    plb_${name}_count[i] = plb_loc_${name}_count;
+                    if (plb_${name}_count)
+                        plb_${name}_count[i] = plb_loc_${name}_count;
                 %endif
             %endfor
         %endif
@@ -725,11 +738,14 @@ class ListOfListsBuilder:
     List entries are generated by calls to `APPEND_<list name>(value)`.
     Multiple lists may be generated at once.
 
+    .. automethod:: __init__
+    .. automethod:: __call__
     """
     def __init__(self, context, list_names_and_dtypes, generate_template,
             arg_decls, count_sharing=None, devices=None,
             name_prefix="plb_build_list", options=[], preamble="",
-            debug=False, complex_kernel=False):
+            debug=False, complex_kernel=False,
+            eliminate_empty_output_lists=False):
         """
         :arg context: A :class:`pyopencl.Context`.
         :arg list_names_and_dtypes: a list of `(name, dtype)` tuples
@@ -803,6 +819,7 @@ class ListOfListsBuilder:
         self.debug = debug
 
         self.complex_kernel = complex_kernel
+        self.eliminate_empty_output_lists = eliminate_empty_output_lists
 
     # {{{ kernel generators
 
@@ -817,6 +834,36 @@ class ListOfListsBuilder:
                 output_statement="ary[i+1] = item;",
                 devices=self.devices)
 
+    @memoize_method
+    def get_compress_kernel(self, index_dtype):
+        arguments = """
+            __global ${index_t} *count,
+            __global ${index_t} *nonempty_indices,
+            __global ${index_t} *compressed_indices,
+            __global ${index_t} *num_non_empty_list
+        """
+        from sys import version_info
+        if (version_info > (3, 0)):
+            arguments = Template(arguments)
+        else:
+            arguments = Template(arguments, disable_unicode=True)
+
+        from pyopencl.scan import GenericScanKernel
+        return GenericScanKernel(
+                self.context, index_dtype,
+                arguments=arguments.render(index_t=dtype_to_ctype(index_dtype)),
+                input_expr="count[i] == 0 ? 0 : 1",
+                scan_expr="a+b", neutral="0",
+                output_statement="""
+                    compressed_indices[i + 1] = item;
+                    if (prev_item != item) {
+                        nonempty_indices[item - 1] = i;
+                        count[item - 1] = count[i];
+                    }
+                    if (i + 1 == N) *num_non_empty_list = item;
+                    """,
+                devices=self.devices)
+
     def do_not_vectorize(self):
         from pytools import any
         return (self.complex_kernel
@@ -851,6 +898,7 @@ class ListOfListsBuilder:
                     self.context.devices),
                 debug=self.debug,
                 do_not_vectorize=self.do_not_vectorize(),
+                eliminate_empty_output_lists=self.eliminate_empty_output_lists,
 
                 kernel_list_arg_decl=_get_arg_decl(kernel_list_args),
                 kernel_list_arg_values=_get_arg_list(user_list_args, prefix="&"),
@@ -901,6 +949,10 @@ class ListOfListsBuilder:
             kernel_list_args.append(
                     VectorArg(index_dtype, "plb_%s_start_index" % name))
 
+            if self.eliminate_empty_output_lists:
+                kernel_list_args.append(
+                    VectorArg(index_dtype, "%s_compressed_indices" % name))
+
             index_name = "plb_%s_index" % name
             user_list_args.append(OtherArg("%s *%s" % (
                 index_ctype, index_name), index_name))
@@ -917,6 +969,7 @@ class ListOfListsBuilder:
                     self.context.devices),
                 debug=self.debug,
                 do_not_vectorize=self.do_not_vectorize(),
+                eliminate_empty_output_lists=self.eliminate_empty_output_lists,
 
                 kernel_list_arg_decl=_get_arg_decl(kernel_list_args),
                 kernel_list_arg_values=kernel_list_arg_values,
@@ -956,6 +1009,11 @@ class ListOfListsBuilder:
             be passed as their :attr:`pyopencl.array.Array.data` attribute instead.
         :arg allocator: optionally, the allocator to use to allocate new
             arrays.
+        :arg omit_lists: An iterable of list names that should *not* be built
+            with this invocation. The kernel code may *not* call ``APPEND_name``
+            for these omitted lists. If it does, undefined behavior will result.
+            The returned *lists* dictionary will not contain an entry for names
+            in *omit_lists*.
         :arg wait_for: |explain-waitfor|
         :returns: a tuple ``(lists, event)``, where
             *lists* a mapping from (built) list names to objects which
@@ -971,7 +1029,22 @@ class ListOfListsBuilder:
 
               This implies that all lists are contiguous.
 
-              *event* is a :class:`pyopencl.Event` for dependency management.
+            If the *eliminate_empty_output_lists* constructor argument is set to
+            True, *lists* has two additional attributes ``num_nonempty_lists`` and
+            ``nonempty_indices``
+
+            * ``num_nonempty_lists`` for the number of nonempty lists.
+            * ``nonempty_indices`` for the index of nonempty list in input objects.
+
+            In this case, `starts` has `num_nonempty_lists` + 1 entries. The *i*'s
+            entry is the start of the *i*'th nonempty list, which is generated by
+            the object with index *nonempty_indices[i]*.
+
+            *event* is a :class:`pyopencl.Event` for dependency management.
+
+        .. versionchanged:: 2016.2
+
+            Added omit_lists.
         """
         if n_objects >= int(np.iinfo(np.int32).max):
             index_dtype = np.int64
@@ -980,10 +1053,15 @@ class ListOfListsBuilder:
         index_dtype = np.dtype(index_dtype)
 
         allocator = kwargs.pop("allocator", None)
+        omit_lists = kwargs.pop("omit_lists", [])
         wait_for = kwargs.pop("wait_for", None)
         if kwargs:
             raise TypeError("invalid keyword arguments: '%s'" % ", ".join(kwargs))
 
+        for oml in omit_lists:
+            if not any(oml == name for name, _ in self.list_names_and_dtypes):
+                raise ValueError("invalid list name '%s' in omit_lists")
+
         result = {}
         count_list_args = []
 
@@ -993,12 +1071,17 @@ class ListOfListsBuilder:
         count_kernel = self.get_count_kernel(index_dtype)
         write_kernel = self.get_write_kernel(index_dtype)
         scan_kernel = self.get_scan_kernel(index_dtype)
+        if self.eliminate_empty_output_lists:
+            compress_kernel = self.get_compress_kernel(index_dtype)
 
         # {{{ allocate memory for counts
 
         for name, dtype in self.list_names_and_dtypes:
             if name in self.count_sharing:
                 continue
+            if name in omit_lists:
+                count_list_args.append(None)
+                continue
 
             counts = cl.array.empty(queue,
                     (n_objects + 1), index_dtype, allocator=allocator)
@@ -1007,7 +1090,12 @@ class ListOfListsBuilder:
 
             # The scan will turn the "counts" array into the "starts" array
             # in-place.
-            result[name] = BuiltList(starts=counts)
+            if self.eliminate_empty_output_lists:
+                result[name] = BuiltList(count=None, starts=counts, lists=None,
+                                         num_nonempty_lists=None,
+                                         nonempty_indices=None)
+            else:
+                result[name] = BuiltList(count=None, starts=counts, lists=None)
             count_list_args.append(counts.data)
 
         # }}}
@@ -1015,7 +1103,7 @@ class ListOfListsBuilder:
         if self.debug:
             gsize = (1,)
             lsize = (1,)
-        elif self.complex_kernel and queue.device.type == cl.device_type.CPU:
+        elif self.do_not_vectorize():
             gsize = (4*queue.device.max_compute_units,)
             lsize = (1,)
         else:
@@ -1026,6 +1114,29 @@ class ListOfListsBuilder:
                 *(tuple(count_list_args) + args + (n_objects,)),
                 **dict(wait_for=wait_for))
 
+        if self.eliminate_empty_output_lists:
+            compress_events = {}
+            for name, dtype in self.list_names_and_dtypes:
+                if name in omit_lists:
+                    continue
+                if name in self.count_sharing:
+                    continue
+
+                info_record = result[name]
+                info_record.nonempty_indices = cl.array.empty(
+                    queue, (n_objects + 1,), index_dtype, allocator=allocator)
+                info_record.num_nonempty_lists = cl.array.empty(
+                    queue, (1,), index_dtype, allocator=allocator)
+                info_record.compressed_indices = cl.array.empty(
+                    queue, (n_objects + 1,), index_dtype, allocator=allocator)
+                info_record.compressed_indices[0] = 0
+                compress_events[name] = compress_kernel(
+                    info_record.starts,
+                    info_record.nonempty_indices,
+                    info_record.compressed_indices,
+                    info_record.num_nonempty_lists,
+                    wait_for=[count_event] + info_record.compressed_indices.events)
+
         # {{{ run scans
 
         scan_events = []
@@ -1033,11 +1144,28 @@ class ListOfListsBuilder:
         for name, dtype in self.list_names_and_dtypes:
             if name in self.count_sharing:
                 continue
+            if name in omit_lists:
+                continue
 
             info_record = result[name]
+            if self.eliminate_empty_output_lists:
+                compress_events[name].wait()
+                num_nonempty_lists = info_record.num_nonempty_lists.get()[0]
+                info_record.num_nonempty_lists = num_nonempty_lists
+                info_record.starts = info_record.starts[:num_nonempty_lists + 1]
+                info_record.nonempty_indices = \
+                    info_record.nonempty_indices[:num_nonempty_lists]
+                info_record.starts[-1] = 0
+
             starts_ary = info_record.starts
-            evt = scan_kernel(starts_ary, wait_for=[count_event],
-                    size=n_objects)
+            if self.eliminate_empty_output_lists:
+                evt = scan_kernel(
+                        starts_ary,
+                        size=info_record.num_nonempty_lists,
+                        wait_for=starts_ary.events)
+            else:
+                evt = scan_kernel(starts_ary, wait_for=[count_event],
+                        size=n_objects)
 
             starts_ary.setitem(0, 0, queue=queue, wait_for=[evt])
             scan_events.extend(starts_ary.events)
@@ -1051,6 +1179,12 @@ class ListOfListsBuilder:
 
         write_list_args = []
         for name, dtype in self.list_names_and_dtypes:
+            if name in omit_lists:
+                write_list_args.append(None)
+                if name not in self.count_sharing:
+                    write_list_args.append(None)
+                continue
+
             if name in self.count_sharing:
                 sharing_from = self.count_sharing[name]
 
@@ -1069,6 +1203,9 @@ class ListOfListsBuilder:
             if name not in self.count_sharing:
                 write_list_args.append(info_record.starts.data)
 
+            if self.eliminate_empty_output_lists:
+                write_list_args.append(info_record.compressed_indices.data)
+
         # }}}
 
         evt = write_kernel(queue, gsize, lsize,
