@@ -1,6 +1,9 @@
 """Elementwise functionality."""
 
 from __future__ import division
+from __future__ import absolute_import
+from six.moves import range
+from six.moves import zip
 
 __copyright__ = "Copyright (C) 2009 Andreas Kloeckner"
 
@@ -33,7 +36,7 @@ import numpy as np
 import pyopencl as cl
 from pytools import memoize_method
 from pyopencl.tools import (dtype_to_ctype, VectorArg, ScalarArg,
-        KernelTemplateBase)
+        KernelTemplateBase, dtype_to_c_struct)
 
 
 # {{{ elementwise kernel code generator
@@ -242,6 +245,7 @@ class ElementwiseKernel:
 
         range_ = kwargs.pop("range", None)
         slice_ = kwargs.pop("slice", None)
+        capture_as = kwargs.pop("capture_as", None)
 
         use_range = range_ is not None or slice_ is not None
         kernel, arg_descrs = self.get_kernel(use_range)
@@ -307,6 +311,12 @@ class ElementwiseKernel:
             invocation_args.append(repr_vec.size)
             gs, ls = repr_vec.get_sizes(queue, max_wg_size)
 
+        if capture_as is not None:
+            kernel.set_args(*invocation_args)
+            kernel.capture_call(
+                    capture_as, queue,
+                    gs, ls, *invocation_args, wait_for=wait_for)
+
         kernel.set_args(*invocation_args)
         return cl.enqueue_nd_range_kernel(queue, kernel,
                 gs, ls, wait_for=wait_for)
@@ -371,7 +381,9 @@ def get_take_kernel(context, dtype, idx_dtype, vec_count=1):
                 "dest%d[i] = src%d[src_idx];" % (i, i)
                 for i in range(vec_count)))
 
-    return get_elwise_kernel(context, args, body, name="take")
+    return get_elwise_kernel(context, args, body,
+            preamble=dtype_to_c_struct(context.devices[0], dtype),
+            name="take")
 
 
 @context_dependent_memoize
@@ -383,7 +395,7 @@ def get_take_put_kernel(context, dtype, idx_dtype, with_offsets, vec_count=1):
 
     args = [
             VectorArg(dtype, "dest%d" % i)
-                for i in range(vec_count)
+            for i in range(vec_count)
             ] + [
                 VectorArg(idx_dtype, "gmem_dest_idx", with_offset=True),
                 VectorArg(idx_dtype, "gmem_src_idx", with_offset=True),
@@ -392,7 +404,7 @@ def get_take_put_kernel(context, dtype, idx_dtype, with_offsets, vec_count=1):
                 for i in range(vec_count)
             ] + [
                 ScalarArg(idx_dtype, "offset%d" % i)
-                    for i in range(vec_count) if with_offsets
+                for i in range(vec_count) if with_offsets
             ]
 
     if with_offsets:
@@ -409,7 +421,9 @@ def get_take_put_kernel(context, dtype, idx_dtype, with_offsets, vec_count=1):
                 "%(idx_tp)s dest_idx = gmem_dest_idx[i];\n" % ctx)
             + "\n".join(get_copy_insn(i) for i in range(vec_count)))
 
-    return get_elwise_kernel(context, args, body, name="take_put")
+    return get_elwise_kernel(context, args, body,
+            preamble=dtype_to_c_struct(context.devices[0], dtype),
+            name="take_put")
 
 
 @context_dependent_memoize
@@ -421,20 +435,30 @@ def get_put_kernel(context, dtype, idx_dtype, vec_count=1):
 
     args = [
             VectorArg(dtype, "dest%d" % i, with_offset=True)
-                for i in range(vec_count)
+            for i in range(vec_count)
             ] + [
                 VectorArg(idx_dtype, "gmem_dest_idx", with_offset=True),
             ] + [
                 VectorArg(dtype, "src%d" % i, with_offset=True)
                 for i in range(vec_count)
+            ] + [
+                VectorArg(np.uint8, "use_fill", with_offset=True)
+            ] + [
+                VectorArg(np.int64, "val_ary_lengths", with_offset=True)
             ]
 
     body = (
             "%(idx_tp)s dest_idx = gmem_dest_idx[i];\n" % ctx
-            + "\n".join("dest%d[dest_idx] = src%d[i];" % (i, i)
-                for i in range(vec_count)))
+            + "\n".join(
+                    "dest{i}[dest_idx] = (use_fill[{i}] ? src{i}[0] : "
+                    "src{i}[i % val_ary_lengths[{i}]]);".format(i=i)
+                    for i in range(vec_count)
+                    )
+            )
 
-    return get_elwise_kernel(context, args, body, name="put")
+    return get_elwise_kernel(context, args, body,
+            preamble=dtype_to_c_struct(context.devices[0], dtype),
+            name="put")
 
 
 @context_dependent_memoize
@@ -446,12 +470,17 @@ def get_copy_kernel(context, dtype_dest, dtype_src):
     if dtype_dest.kind == "c" and dtype_src != dtype_dest:
         src = "%s_cast(%s)" % (complex_dtype_to_name(dtype_dest), src),
 
+    if dtype_dest != dtype_src and (
+            dtype_dest.kind == "V" or dtype_src.kind == "V"):
+        raise TypeError("copying between non-identical struct types")
+
     return get_elwise_kernel(context,
             "%(tp_dest)s *dest, %(tp_src)s *src" % {
                 "tp_dest": dtype_to_ctype(dtype_dest),
                 "tp_src": dtype_to_ctype(dtype_src),
                 },
             "dest[i] = %s" % src,
+            preamble=dtype_to_c_struct(context.devices[0], dtype_dest),
             name="copy")
 
 
@@ -525,7 +554,6 @@ def get_axpbyz_kernel(context, dtype_x, dtype_y, dtype_z):
 
     x_is_complex = dtype_x.kind == "c"
     y_is_complex = dtype_y.kind == "c"
-    z_is_complex = dtype_z.kind == "c"
 
     if x_is_complex:
         ax = "%s_mul(a, x[i])" % complex_dtype_to_name(dtype_x)
@@ -539,9 +567,15 @@ def get_axpbyz_kernel(context, dtype_x, dtype_y, dtype_z):
     if not x_is_complex and y_is_complex:
         ax = "%s_fromreal(%s)" % (complex_dtype_to_name(dtype_y), ax)
 
-    result = "%s + %s" % (ax, by)
-    if z_is_complex:
-        result = "%s_cast(%s)" % (complex_dtype_to_name(dtype_z), result)
+    if x_is_complex or y_is_complex:
+        result = (
+                "{root}_add({root}_cast({ax}), {root}_cast({by}))"
+                .format(
+                    ax=ax,
+                    by=by,
+                    root=complex_dtype_to_name(dtype_z)))
+    else:
+        result = "%s + %s" % (ax, by)
 
     return get_elwise_kernel(context,
             "%(tp_z)s *z, %(tp_x)s a, %(tp_x)s *x, %(tp_y)s b, %(tp_y)s *y" % {
@@ -562,24 +596,27 @@ def get_axpbz_kernel(context, dtype_a, dtype_x, dtype_b, dtype_z):
     z_is_complex = dtype_z.kind == "c"
 
     ax = "a*x[i]"
-    if a_is_complex and x_is_complex:
+    if x_is_complex:
+        a = "a"
+        x = "x[i]"
+
+        if dtype_x != dtype_z:
+            x = "%s_cast(%s)" % (complex_dtype_to_name(dtype_z), x)
+
+        if a_is_complex:
+            if dtype_a != dtype_z:
+                a = "%s_cast(%s)" % (complex_dtype_to_name(dtype_z), a)
+
+            ax = "%s_mul(%s, %s)" % (complex_dtype_to_name(dtype_z), a, x)
+        else:
+            ax = "%s_rmul(%s, %s)" % (complex_dtype_to_name(dtype_z), a, x)
+    elif a_is_complex:
         a = "a"
         x = "x[i]"
 
         if dtype_a != dtype_z:
             a = "%s_cast(%s)" % (complex_dtype_to_name(dtype_z), a)
-        if dtype_x != dtype_z:
-            x = "%s_cast(%s)" % (complex_dtype_to_name(dtype_z), x)
-
-        ax = "%s_mul(%s, %s)" % (complex_dtype_to_name(dtype_z), a, x)
-
-    # The following two are workarounds for Apple on OS X 10.8.
-    # They're not really necessary.
-
-    elif a_is_complex and not x_is_complex:
-        ax = "a*((%s) x[i])" % dtype_to_ctype(real_dtype(dtype_a))
-    elif not a_is_complex and x_is_complex:
-        ax = "((%s) a)*x[i]" % dtype_to_ctype(real_dtype(dtype_x))
+        ax = "%s_mulr(%s, %s)" % (complex_dtype_to_name(dtype_z), a, x)
 
     b = "b"
     if z_is_complex and not b_is_complex:
@@ -592,6 +629,14 @@ def get_axpbz_kernel(context, dtype_a, dtype_x, dtype_b, dtype_z):
         ax = "%s_cast(%s)" % (complex_dtype_to_name(dtype_z), ax)
         b = "%s_cast(%s)" % (complex_dtype_to_name(dtype_z), b)
 
+    if a_is_complex or x_is_complex or b_is_complex:
+        expr = "{root}_add({ax}, {b})".format(
+                ax=ax,
+                b=b,
+                root=complex_dtype_to_name(dtype_z))
+    else:
+        expr = "%s + %s" % (ax, b)
+
     return get_elwise_kernel(context,
             "%(tp_z)s *z, %(tp_a)s a, %(tp_x)s *x,%(tp_b)s b" % {
                 "tp_a": dtype_to_ctype(dtype_a),
@@ -599,7 +644,7 @@ def get_axpbz_kernel(context, dtype_a, dtype_x, dtype_b, dtype_z):
                 "tp_b": dtype_to_ctype(dtype_b),
                 "tp_z": dtype_to_ctype(dtype_z),
                 },
-            "z[i] = %s + %s" % (ax, b),
+            "z[i] = " + expr,
             name="axpb")
 
 
@@ -607,7 +652,6 @@ def get_axpbz_kernel(context, dtype_a, dtype_x, dtype_b, dtype_z):
 def get_multiply_kernel(context, dtype_x, dtype_y, dtype_z):
     x_is_complex = dtype_x.kind == "c"
     y_is_complex = dtype_y.kind == "c"
-    z_is_complex = dtype_z.kind == "c"
 
     x = "x[i]"
     y = "y[i]"
@@ -619,13 +663,13 @@ def get_multiply_kernel(context, dtype_x, dtype_y, dtype_z):
 
     if x_is_complex and y_is_complex:
         xy = "%s_mul(%s, %s)" % (complex_dtype_to_name(dtype_z), x, y)
-
+    elif x_is_complex and not y_is_complex:
+        xy = "%s_mulr(%s, %s)" % (complex_dtype_to_name(dtype_z), x, y)
+    elif not x_is_complex and y_is_complex:
+        xy = "%s_rmul(%s, %s)" % (complex_dtype_to_name(dtype_z), x, y)
     else:
         xy = "%s * %s" % (x, y)
 
-    if z_is_complex:
-        xy = "%s_cast(%s)" % (complex_dtype_to_name(dtype_z), xy)
-
     return get_elwise_kernel(context,
             "%(tp_z)s *z, %(tp_x)s *x, %(tp_y)s *y" % {
                 "tp_x": dtype_to_ctype(dtype_x),
@@ -654,8 +698,9 @@ def get_divide_kernel(context, dtype_x, dtype_y, dtype_z):
     if x_is_complex and y_is_complex:
         xoy = "%s_divide(%s, %s)" % (complex_dtype_to_name(dtype_z), x, y)
     elif not x_is_complex and y_is_complex:
-
         xoy = "%s_rdivide(%s, %s)" % (complex_dtype_to_name(dtype_z), x, y)
+    elif x_is_complex and not y_is_complex:
+        xoy = "%s_divider(%s, %s)" % (complex_dtype_to_name(dtype_z), x, y)
     else:
         xoy = "%s / %s" % (x, y)
 
@@ -691,7 +736,9 @@ def get_rdivide_elwise_kernel(context, dtype_x, dtype_y, dtype_z):
     if x_is_complex and y_is_complex:
         yox = "%s_divide(%s, %s)" % (complex_dtype_to_name(dtype_z), y, x)
     elif not y_is_complex and x_is_complex:
-        yox = "%s_rdivide(%s, %s)" % (complex_dtype_to_name(dtype_x), y, x)
+        yox = "%s_rdivide(%s, %s)" % (complex_dtype_to_name(dtype_z), y, x)
+    elif y_is_complex and not x_is_complex:
+        yox = "%s_divider(%s, %s)" % (complex_dtype_to_name(dtype_z), y, x)
     else:
         yox = "%s / %s" % (y, x)
 
@@ -712,6 +759,7 @@ def get_fill_kernel(context, dtype):
                 "tp": dtype_to_ctype(dtype),
                 },
             "z[i] = a",
+            preamble=dtype_to_c_struct(context.devices[0], dtype),
             name="fill")
 
 
@@ -728,16 +776,18 @@ def get_reverse_kernel(context, dtype):
 @context_dependent_memoize
 def get_arange_kernel(context, dtype):
     if dtype.kind == "c":
-        i = "%s_fromreal(i)" % complex_dtype_to_name(dtype)
+        expr = (
+                "{root}_add(start, {root}_rmul(i, step))"
+                .format(root=complex_dtype_to_name(dtype)))
     else:
-        i = "(%s) i" % dtype_to_ctype(dtype)
+        expr = "start + ((%s) i)*step" % dtype_to_ctype(dtype)
 
     return get_elwise_kernel(context, [
         VectorArg(dtype, "z", with_offset=True),
         ScalarArg(dtype, "start"),
         ScalarArg(dtype, "step"),
         ],
-        "z[i] = start + %s*step" % i,
+        "z[i] = " + expr,
         name="arange")
 
 
@@ -792,6 +842,38 @@ def get_pow_kernel(context, dtype_x, dtype_y, dtype_z,
             name="pow_method")
 
 
+@context_dependent_memoize
+def get_unop_kernel(context, operator, res_dtype, in_dtype):
+    return get_elwise_kernel(context, [
+        VectorArg(res_dtype, "z", with_offset=True),
+        VectorArg(in_dtype, "y", with_offset=True),
+        ],
+        "z[i] = %s y[i]" % operator,
+        name="unary_op_kernel")
+
+
+@context_dependent_memoize
+def get_array_scalar_binop_kernel(context, operator, dtype_res, dtype_a, dtype_b):
+    return get_elwise_kernel(context, [
+        VectorArg(dtype_res, "out", with_offset=True),
+        VectorArg(dtype_a, "a", with_offset=True),
+        ScalarArg(dtype_b, "b"),
+        ],
+        "out[i] = a[i] %s b" % operator,
+        name="scalar_binop_kernel")
+
+
+@context_dependent_memoize
+def get_array_binop_kernel(context, operator, dtype_res, dtype_a, dtype_b):
+    return get_elwise_kernel(context, [
+        VectorArg(dtype_res, "out", with_offset=True),
+        VectorArg(dtype_a, "a", with_offset=True),
+        VectorArg(dtype_b, "b", with_offset=True),
+        ],
+        "out[i] = a[i] %s b[i]" % operator,
+        name="binop_kernel")
+
+
 @context_dependent_memoize
 def get_array_scalar_comparison_kernel(context, operator, dtype_a):
     return get_elwise_kernel(context, [
@@ -912,20 +994,80 @@ def get_ldexp_kernel(context, out_dtype=np.float32, sig_dtype=np.float32,
 @context_dependent_memoize
 def get_bessel_kernel(context, which_func, out_dtype=np.float64,
                       order_dtype=np.int32, x_dtype=np.float64):
+    if x_dtype.kind != "c":
+        return get_elwise_kernel(context, [
+            VectorArg(out_dtype, "z", with_offset=True),
+            ScalarArg(order_dtype, "ord_n"),
+            VectorArg(x_dtype, "x", with_offset=True),
+            ],
+            "z[i] = bessel_%sn(ord_n, x[i])" % which_func,
+            name="bessel_%sn_kernel" % which_func,
+            preamble="""
+            #if __OPENCL_C_VERSION__ < 120
+            #pragma OPENCL EXTENSION cl_khr_fp64: enable
+            #endif
+            #define PYOPENCL_DEFINE_CDOUBLE
+            #include <pyopencl-bessel-%s.cl>
+            """ % which_func)
+    else:
+        if which_func != "j":
+            raise NotImplementedError("complex arguments for Bessel Y")
+
+        if x_dtype != np.complex128:
+            raise NotImplementedError("non-complex double dtype")
+        if x_dtype != out_dtype:
+            raise NotImplementedError("different input/output types")
+
+        return get_elwise_kernel(context, [
+            VectorArg(out_dtype, "z", with_offset=True),
+            ScalarArg(order_dtype, "ord_n"),
+            VectorArg(x_dtype, "x", with_offset=True),
+            ],
+            """
+            cdouble_t jv_loc;
+            cdouble_t jvp1_loc;
+            bessel_j_complex(ord_n, x[i], &jv_loc, &jvp1_loc);
+            z[i] = jv_loc;
+            """,
+            name="bessel_j_complex_kernel",
+            preamble="""
+            #if __OPENCL_C_VERSION__ < 120
+            #pragma OPENCL EXTENSION cl_khr_fp64: enable
+            #endif
+            #define PYOPENCL_DEFINE_CDOUBLE
+            #include <pyopencl-complex.h>
+            #include <pyopencl-bessel-j-complex.cl>
+            """)
+
+
+@context_dependent_memoize
+def get_hankel_01_kernel(context, out_dtype, x_dtype):
+    if x_dtype != np.complex128:
+        raise NotImplementedError("non-complex double dtype")
+    if x_dtype != out_dtype:
+        raise NotImplementedError("different input/output types")
+
     return get_elwise_kernel(context, [
-        VectorArg(out_dtype, "z", with_offset=True),
-        ScalarArg(order_dtype, "ord_n"),
+        VectorArg(out_dtype, "h0", with_offset=True),
+        VectorArg(out_dtype, "h1", with_offset=True),
         VectorArg(x_dtype, "x", with_offset=True),
         ],
-        "z[i] = bessel_%sn(ord_n, x[i])" % which_func,
-        name="bessel_%sn_kernel" % which_func,
+        """
+        cdouble_t h0_loc;
+        cdouble_t h1_loc;
+        hankel_01_complex(x[i], &h0_loc, &h1_loc, 1);
+        h0[i] = h0_loc;
+        h1[i] = h1_loc;
+        """,
+        name="hankel_complex_kernel",
         preamble="""
         #if __OPENCL_C_VERSION__ < 120
         #pragma OPENCL EXTENSION cl_khr_fp64: enable
         #endif
         #define PYOPENCL_DEFINE_CDOUBLE
-        #include <pyopencl-bessel-%s.cl>
-        """ % which_func)
+        #include <pyopencl-complex.h>
+        #include <pyopencl-hankel-complex.cl>
+        """)
 
 
 @context_dependent_memoize
