@@ -1,6 +1,6 @@
 """Scan primitive."""
 
-from __future__ import division
+from __future__ import division, absolute_import
 
 __copyright__ = """
 Copyright 2011-2012 Andreas Kloeckner
@@ -20,13 +20,12 @@ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 
-Derived from thrust/detail/backend/cuda/detail/fast_scan.inl
-within the Thrust project, https://code.google.com/p/thrust/
+Derived from code within the Thrust project, https://github.com/thrust/thrust/
 
 """
 
-# Direct link to thrust source:
-# https://code.google.com/p/thrust/source/browse/thrust/detail/backend/cuda/detail/fast_scan.inl # noqa
+import six
+from six.moves import range, zip
 
 import numpy as np
 
@@ -35,11 +34,18 @@ import pyopencl.array  # noqa
 from pyopencl.tools import (dtype_to_ctype, bitlog2,
         KernelTemplateBase, _process_code_for_macro,
         get_arg_list_scalar_arg_dtypes,
-        context_dependent_memoize)
+        context_dependent_memoize,
+        _NumpyTypesKeyBuilder)
 
 import pyopencl._mymako as mako
 from pyopencl._cluda import CLUDA_PREAMBLE
 
+from pytools.persistent_dict import WriteOncePersistentDict
+
+
+import logging
+logger = logging.getLogger(__name__)
+
 
 # {{{ preamble
 
@@ -296,8 +302,11 @@ void ${kernel_name}(
 
             if (LID_0 == 0 && unit_base != interval_begin)
             {
+                scan_type tmp = ldata[K][WG_SIZE - 1].value;
+                scan_type tmp_aux = ldata[0][0].value;
+
                 ldata[0][0].value = SCAN_EXPR(
-                    ldata[K][WG_SIZE - 1].value, ldata[0][0].value,
+                    tmp, tmp_aux,
                     %if is_segmented:
                         (l_segment_start_flags[0][0])
                     %else:
@@ -742,6 +751,7 @@ def _round_down_to_power_of_2(val):
     assert result <= val
     return result
 
+
 _PREFIX_WORDS = set("""
         ldata partial_scan_buffer global scan_offset
         segment_start_in_k_group carry
@@ -760,7 +770,7 @@ _PREFIX_WORDS = set("""
         group_base seg_end my_val DEBUG ARGS
         ints_to_store ints_per_wg scan_types_per_int linear_index
         linear_scan_data_idx dest src store_base wrapped_scan_type
-        dummy
+        dummy scan_tmp tmp_aux
 
         LID_2 LID_1 LID_0
         LDIM_0 LDIM_1 LDIM_2
@@ -850,11 +860,72 @@ def _make_template(s):
 
     return mako.template.Template(s, strict_undefined=True)
 
-from pytools import Record
+
+from pytools import Record, RecordWithoutPickling
 
 
-class _ScanKernelInfo(Record):
-    pass
+class _GeneratedScanKernelInfo(Record):
+
+    __slots__ = [
+            "scan_src",
+            "kernel_name",
+            "scalar_arg_dtypes",
+            "wg_size",
+            "k_group_size"]
+
+    def __init__(self, scan_src, kernel_name, scalar_arg_dtypes, wg_size,
+            k_group_size):
+        Record.__init__(self,
+                scan_src=scan_src,
+                kernel_name=kernel_name,
+                scalar_arg_dtypes=scalar_arg_dtypes,
+                wg_size=wg_size,
+                k_group_size=k_group_size)
+
+    def build(self, context, options):
+        program = cl.Program(context, self.scan_src).build(options)
+        kernel = getattr(program, self.kernel_name)
+        kernel.set_scalar_arg_dtypes(self.scalar_arg_dtypes)
+        return _BuiltScanKernelInfo(
+                kernel=kernel,
+                wg_size=self.wg_size,
+                k_group_size=self.k_group_size)
+
+
+class _BuiltScanKernelInfo(RecordWithoutPickling):
+
+    __slots__ = ["kernel", "wg_size", "k_group_size"]
+
+    def __init__(self, kernel, wg_size, k_group_size):
+        RecordWithoutPickling.__init__(self,
+                kernel=kernel,
+                wg_size=wg_size,
+                k_group_size=k_group_size)
+
+
+class _GeneratedFinalUpdateKernelInfo(Record):
+
+    def __init__(self, source, kernel_name, scalar_arg_dtypes, update_wg_size):
+        Record.__init__(self,
+                source=source,
+                kernel_name=kernel_name,
+                scalar_arg_dtypes=scalar_arg_dtypes,
+                update_wg_size=update_wg_size)
+
+    def build(self, context, options):
+        program = cl.Program(context, self.source).build(options)
+        kernel = getattr(program, self.kernel_name)
+        kernel.set_scalar_arg_dtypes(self.scalar_arg_dtypes)
+        return _BuiltFinalUpdateKernelInfo(kernel, self.update_wg_size)
+
+
+class _BuiltFinalUpdateKernelInfo(RecordWithoutPickling):
+    __slots__ = ["kernel", "update_wg_size"]
+
+    def __init__(self, kernel, update_wg_size):
+        RecordWithoutPickling.__init__(self,
+                kernel=kernel,
+                update_wg_size=update_wg_size)
 
 # }}}
 
@@ -1027,13 +1098,45 @@ class _GenericScanKernelBase(object):
                 has_double_support(dev) for dev in devices),
             )
 
+        index_typename = dtype_to_ctype(self.index_dtype)
+        scan_typename = dtype_to_ctype(dtype)
+
+        # This key is meant to uniquely identify the non-device parameters for
+        # the scan kernel.
+        self.kernel_key = (
+            self.dtype,
+            tuple(arg.declarator() for arg in self.parsed_args),
+            self.input_expr,
+            scan_expr,
+            neutral,
+            output_statement,
+            is_segment_start_expr,
+            tuple(input_fetch_exprs),
+            index_dtype,
+            name_prefix,
+            preamble,
+            # These depend on dtype_to_ctype(), so their value is independent of
+            # the other variables.
+            index_typename,
+            scan_typename,
+            )
+
         # }}}
 
+        self.use_lookbehind_update = "prev_item" in self.output_statement
+        self.store_segment_start_flags = (
+                self.is_segmented and self.use_lookbehind_update)
+
         self.finish_setup()
 
     # }}}
 
 
+generic_scan_kernel_cache = WriteOncePersistentDict(
+        "pyopencl-generated-scan-kernel-cache-v1",
+        key_builder=_NumpyTypesKeyBuilder())
+
+
 class GenericScanKernel(_GenericScanKernelBase):
     """Generates and executes code that performs prefix sums ("scans") on
     arbitrary types, with many possible tweaks.
@@ -1049,14 +1152,57 @@ class GenericScanKernel(_GenericScanKernelBase):
                 output_statement="ary[i+1] = item;")
 
         a = cl.array.arange(queue, 10000, dtype=np.int32)
-        scan_kernel(a, queue=queue)
+        knl(a, queue=queue)
 
     """
 
     def finish_setup(self):
-        use_lookbehind_update = "prev_item" in self.output_statement
-        self.store_segment_start_flags = self.is_segmented and use_lookbehind_update
+        # Before generating the kernel, see if it's cached.
+        from pyopencl.cache import get_device_cache_id
+        devices_key = tuple(get_device_cache_id(device)
+                for device in self.devices)
 
+        cache_key = (self.kernel_key, devices_key)
+
+        from_cache = False
+
+        try:
+            result = generic_scan_kernel_cache[cache_key]
+            from_cache = True
+            logger.debug(
+                    "cache hit for generated scan kernel '%s'" % self.name_prefix)
+            (
+                self.first_level_scan_gen_info,
+                self.second_level_scan_gen_info,
+                self.final_update_gen_info) = result
+        except KeyError:
+            pass
+
+        if not from_cache:
+            logger.debug(
+                    "cache miss for generated scan kernel '%s'" % self.name_prefix)
+            self._finish_setup_impl()
+
+            result = (self.first_level_scan_gen_info,
+                      self.second_level_scan_gen_info,
+                      self.final_update_gen_info)
+
+            generic_scan_kernel_cache.store_if_not_present(cache_key, result)
+
+        # Build the kernels.
+        self.first_level_scan_info = self.first_level_scan_gen_info.build(
+                self.context, self.options)
+        del self.first_level_scan_gen_info
+
+        self.second_level_scan_info = self.second_level_scan_gen_info.build(
+                self.context, self.options)
+        del self.second_level_scan_gen_info
+
+        self.final_update_info = self.final_update_gen_info.build(
+                self.context, self.options)
+        del self.final_update_gen_info
+
+    def _finish_setup_impl(self):
         # {{{ find usable workgroup/k-group size, build first-level scan
 
         trip_count = 0
@@ -1065,6 +1211,10 @@ class GenericScanKernel(_GenericScanKernelBase):
                 dev.local_mem_size
                 for dev in self.devices)
 
+        if "CUDA" in self.devices[0].platform.name:
+            # not sure where these go, but roughly this much seems unavailable.
+            avail_local_mem -= 0x400
+
         is_cpu = self.devices[0].type & cl.device_type.CPU
         is_gpu = self.devices[0].type & cl.device_type.GPU
 
@@ -1091,7 +1241,7 @@ class GenericScanKernel(_GenericScanKernelBase):
                 k_group_size = 2**k_exp
                 lmem_use = self.get_local_mem_use(wg_size, k_group_size,
                         use_bank_conflict_avoidance)
-                if lmem_use + 256 <= avail_local_mem:
+                if lmem_use <= avail_local_mem:
                     solutions.append((wg_size*k_group_size, k_group_size, wg_size))
 
         if is_gpu:
@@ -1110,7 +1260,7 @@ class GenericScanKernel(_GenericScanKernelBase):
         _, k_group_size, max_scan_wg_size = max(solutions)
 
         while True:
-            candidate_scan_info = self.build_scan_kernel(
+            candidate_scan_gen_info = self.generate_scan_kernel(
                     max_scan_wg_size, self.parsed_args,
                     _process_code_for_macro(self.input_expr),
                     self.is_segment_start_expr,
@@ -1120,6 +1270,9 @@ class GenericScanKernel(_GenericScanKernelBase):
                     k_group_size=k_group_size,
                     use_bank_conflict_avoidance=use_bank_conflict_avoidance)
 
+            candidate_scan_info = candidate_scan_gen_info.build(
+                    self.context, self.options)
+
             # Will this device actually let us execute this kernel
             # at the desired work group size? Building it is the
             # only way to find out.
@@ -1137,7 +1290,7 @@ class GenericScanKernel(_GenericScanKernelBase):
             trip_count += 1
             assert trip_count <= 20
 
-        self.first_level_scan_info = candidate_scan_info
+        self.first_level_scan_gen_info = candidate_scan_gen_info
         assert (_round_down_to_power_of_2(candidate_scan_info.wg_size)
                 == candidate_scan_info.wg_size)
 
@@ -1165,7 +1318,7 @@ class GenericScanKernel(_GenericScanKernelBase):
         else:
             second_level_build_kwargs["is_segment_start_expr"] = None
 
-        self.second_level_scan_info = self.build_scan_kernel(
+        self.second_level_scan_gen_info = self.generate_scan_kernel(
                 max_scan_wg_size,
                 arguments=second_level_arguments,
                 input_expr="interval_sums[i]",
@@ -1178,26 +1331,21 @@ class GenericScanKernel(_GenericScanKernelBase):
 
         # }}}
 
-        # {{{ build final update kernel
+        # {{{ generate final update kernel
 
-        self.update_wg_size = min(max_scan_wg_size, 256)
+        update_wg_size = min(max_scan_wg_size, 256)
 
         final_update_tpl = _make_template(UPDATE_SOURCE)
         final_update_src = str(final_update_tpl.render(
-            wg_size=self.update_wg_size,
+            wg_size=update_wg_size,
             output_statement=self.output_statement,
             argument_signature=", ".join(
                 arg.declarator() for arg in self.parsed_args),
             is_segment_start_expr=self.is_segment_start_expr,
             input_expr=_process_code_for_macro(self.input_expr),
-            use_lookbehind_update=use_lookbehind_update,
+            use_lookbehind_update=self.use_lookbehind_update,
             **self.code_variables))
 
-        final_update_prg = cl.Program(
-                self.context, final_update_src).build(self.options)
-        self.final_update_knl = getattr(
-                final_update_prg,
-                self.name_prefix+"_final_update")
         update_scalar_arg_dtypes = (
                 get_arg_list_scalar_arg_dtypes(self.parsed_args)
                 + [self.index_dtype, self.index_dtype, None, None])
@@ -1206,7 +1354,12 @@ class GenericScanKernel(_GenericScanKernelBase):
             update_scalar_arg_dtypes.append(None)
         if self.store_segment_start_flags:
             update_scalar_arg_dtypes.append(None)  # g_segment_start_flags
-        self.final_update_knl.set_scalar_arg_dtypes(update_scalar_arg_dtypes)
+
+        self.final_update_gen_info = _GeneratedFinalUpdateKernelInfo(
+                final_update_src,
+                self.name_prefix + "_final_update",
+                update_scalar_arg_dtypes,
+                update_wg_size)
 
         # }}}
 
@@ -1237,10 +1390,10 @@ class GenericScanKernel(_GenericScanKernelBase):
 
                 + k_group_size*wg_size*sum(
                     arg_dtypes[arg_name].itemsize
-                    for arg_name, ife_offsets in fetch_expr_offsets.items()
+                    for arg_name, ife_offsets in list(fetch_expr_offsets.items())
                     if -1 in ife_offsets or len(ife_offsets) > 1))
 
-    def build_scan_kernel(self, max_wg_size, arguments, input_expr,
+    def generate_scan_kernel(self, max_wg_size, arguments, input_expr,
             is_segment_start_expr, input_fetch_exprs, is_first_level,
             store_segment_start_flags, k_group_size,
             use_bank_conflict_avoidance):
@@ -1250,7 +1403,7 @@ class GenericScanKernel(_GenericScanKernelBase):
         wg_size = _round_down_to_power_of_2(
                 min(max_wg_size, 256))
 
-        kernel_name = self.code_variables["name_prefix"]+"_scan_intervals"
+        kernel_name = self.code_variables["name_prefix"]
         if is_first_level:
             kernel_name += "_lev1"
         else:
@@ -1270,22 +1423,21 @@ class GenericScanKernel(_GenericScanKernelBase):
             kernel_name=kernel_name,
             **self.code_variables))
 
-        prg = cl.Program(self.context, scan_src).build(self.options)
-
-        knl = getattr(prg, kernel_name)
-
         scalar_arg_dtypes.extend(
-                (None, self.index_dtype, self. index_dtype))
+                (None, self.index_dtype, self.index_dtype))
         if is_first_level:
             scalar_arg_dtypes.append(None)  # interval_results
         if self.is_segmented and is_first_level:
             scalar_arg_dtypes.append(None)  # g_first_segment_start_in_interval
         if store_segment_start_flags:
             scalar_arg_dtypes.append(None)  # g_segment_start_flags
-        knl.set_scalar_arg_dtypes(scalar_arg_dtypes)
 
-        return _ScanKernelInfo(
-                kernel=knl, wg_size=wg_size, knl=knl, k_group_size=k_group_size)
+        return _GeneratedScanKernelInfo(
+                scan_src=scan_src,
+                kernel_name=kernel_name,
+                scalar_arg_dtypes=scalar_arg_dtypes,
+                wg_size=wg_size,
+                k_group_size=k_group_size)
 
     # }}}
 
@@ -1400,14 +1552,16 @@ class GenericScanKernel(_GenericScanKernelBase):
         if self.store_segment_start_flags:
             upd_args.append(segment_start_flags.data)
 
-        return self.final_update_knl(
-                queue, (num_intervals,), (self.update_wg_size,),
+        return self.final_update_info.kernel(
+                queue, (num_intervals,),
+                (self.final_update_info.update_wg_size,),
                 *upd_args, **dict(g_times_l=True, wait_for=[l2_evt]))
 
         # }}}
 
 # }}}
 
+
 # {{{ debug kernel
 
 DEBUG_SCAN_TEMPLATE = SHARED_PREAMBLE + r"""//CL//
@@ -1415,11 +1569,12 @@ DEBUG_SCAN_TEMPLATE = SHARED_PREAMBLE + r"""//CL//
 KERNEL
 REQD_WG_SIZE(1, 1, 1)
 void ${name_prefix}_debug_scan(
+    __global scan_type *scan_tmp,
     ${argument_signature},
     const index_type N)
 {
-    scan_type item = ${neutral};
-    scan_type prev_item;
+    scan_type current = ${neutral};
+    scan_type prev;
 
     for (index_type i = 0; i < N; ++i)
     {
@@ -1427,7 +1582,7 @@ void ${name_prefix}_debug_scan(
             ${arg_ctypes[arg_name]} ${name};
             %if ife_offset < 0:
                 if (i+${ife_offset} >= 0)
-                    ${name} = ${arg_name}[i+offset];
+                    ${name} = ${arg_name}[i+${ife_offset}];
             %else:
                 ${name} = ${arg_name}[i];
             %endif
@@ -1435,18 +1590,31 @@ void ${name_prefix}_debug_scan(
 
         scan_type my_val = INPUT_EXPR(i);
 
-        prev_item = item;
+        prev = current;
         %if is_segmented:
             bool is_seg_start = IS_SEG_START(i, my_val);
         %endif
 
-        item = SCAN_EXPR(prev_item, my_val,
+        current = SCAN_EXPR(prev, my_val,
             %if is_segmented:
                 is_seg_start
             %else:
                 false
             %endif
             );
+        scan_tmp[i] = current;
+    }
+
+    scan_type last_item = scan_tmp[N-1];
+
+    for (index_type i = 0; i < N; ++i)
+    {
+        scan_type item = scan_tmp[i];
+        scan_type prev_item;
+        if (i)
+            prev_item = scan_tmp[i-1];
+        else
+            prev_item = ${neutral};
 
         {
             ${output_statement};
@@ -1473,7 +1641,8 @@ class GenericDebugScanKernel(_GenericScanKernelBase):
         self.kernel = getattr(
                 scan_prg, self.name_prefix+"_debug_scan")
         scalar_arg_dtypes = (
-                get_arg_list_scalar_arg_dtypes(self.parsed_args)
+                [None]
+                + get_arg_list_scalar_arg_dtypes(self.parsed_args)
                 + [self.index_dtype])
         self.kernel.set_scalar_arg_dtypes(scalar_arg_dtypes)
 
@@ -1496,7 +1665,11 @@ class GenericDebugScanKernel(_GenericScanKernelBase):
         if n is None:
             n, = first_array.shape
 
-        data_args = []
+        scan_tmp = cl.array.empty(queue,
+                n, dtype=self.dtype,
+                allocator=allocator)
+
+        data_args = [scan_tmp.data]
         from pyopencl.tools import VectorArg
         for arg_descr, arg_val in zip(self.parsed_args, args):
             if isinstance(arg_descr, VectorArg):
@@ -1536,7 +1709,7 @@ class _LegacyScanKernelBase(GenericScanKernel):
         if output_ary is None:
             output_ary = input_ary
 
-        if isinstance(output_ary, (str, unicode)) and output_ary == "new":
+        if isinstance(output_ary, (str, six.text_type)) and output_ary == "new":
             output_ary = cl.array.empty_like(input_ary, allocator=allocator)
 
         if input_ary.shape != output_ary.shape:
