@@ -1,6 +1,6 @@
 """Various helpful bits and pieces without much of a common theme."""
 
-from __future__ import division
+from __future__ import division, absolute_import
 
 __copyright__ = "Copyright (C) 2010 Andreas Kloeckner"
 
@@ -28,10 +28,15 @@ OTHER DEALINGS IN THE SOFTWARE.
 """
 
 
+import six
+from six.moves import zip, intern
+
 import numpy as np
 from decorator import decorator
 import pyopencl as cl
 from pytools import memoize, memoize_method
+from pyopencl.cffi_cl import _lib
+from pytools.persistent_dict import KeyBuilder as KeyBuilderBase
 
 import re
 
@@ -41,45 +46,23 @@ from pyopencl.compyte.dtypes import (  # noqa
 
 
 def _register_types():
-    from pyopencl.compyte.dtypes import _fill_dtype_registry
-    import struct
+    from pyopencl.compyte.dtypes import (
+            TYPE_REGISTRY, fill_registry_with_opencl_c_types)
 
-    _fill_dtype_registry(respect_windows=False, include_bool=False)
+    fill_registry_with_opencl_c_types(TYPE_REGISTRY)
 
     get_or_register_dtype("cfloat_t", np.complex64)
     get_or_register_dtype("cdouble_t", np.complex128)
 
-    is_64_bit = struct.calcsize('@P') * 8 == 64
-    if not is_64_bit:
-        get_or_register_dtype(
-                ["unsigned long", "unsigned long int"], np.uint64)
-        get_or_register_dtype(
-                ["signed long", "signed long int", "long int"], np.int64)
 
 _register_types()
 
 
 # {{{ imported names
 
-bitlog2 = cl.bitlog2
-
-PooledBuffer = cl.PooledBuffer
-
-from pyopencl._cl import _tools_DeferredAllocator as DeferredAllocator
-from pyopencl._cl import (  # noqa
-        _tools_ImmediateAllocator as ImmediateAllocator)
-
-
-class CLAllocator(DeferredAllocator):
-    def __init__(self, *args, **kwargs):
-        from warnings import warn
-        warn("pyopencl.tools.CLAllocator is deprecated. "
-                "It will be continue to exist throughout the 2013.x "
-                "versions of PyOpenCL. Use {Deferred,Immediate}Allocator.",
-                DeprecationWarning, 2)
-        DeferredAllocator.__init__(self, *args, **kwargs)
-
-MemoryPool = cl.MemoryPool
+bitlog2 = _lib.bitlog2
+from pyopencl.mempool import (  # noqa
+        PooledBuffer, DeferredAllocator, ImmediateAllocator, MemoryPool)
 
 # }}}
 
@@ -116,6 +99,7 @@ def first_arg_dependent_memoize(func, cl_object, *args):
         arg_dict[args] = result
         return result
 
+
 context_dependent_memoize = first_arg_dependent_memoize
 
 
@@ -134,8 +118,8 @@ def first_arg_dependent_memoize_nested(nested_func):
 
     from functools import wraps
     cache_dict_name = intern("_memoize_inner_dic_%s_%s_%d"
-            % (nested_func.__name__, nested_func.func_code.co_filename,
-                nested_func.func_code.co_firstlineno))
+            % (nested_func.__name__, nested_func.__code__.co_filename,
+                nested_func.__code__.co_firstlineno))
 
     from inspect import currentframe
     # prevent ref cycle
@@ -178,6 +162,7 @@ def clear_first_arg_caches():
     for cache in _first_arg_dependent_caches:
         cache.clear()
 
+
 import atexit
 atexit.register(clear_first_arg_caches)
 
@@ -257,7 +242,10 @@ def pytest_generate_tests_for_pyopencl(metafunc):
             return cl.Context([self.device])
 
         def __str__(self):
-            return "<context factory for %s>" % self.device
+            # Don't show address, so that parallel test collection works
+            return ("<context factory for <pyopencl.Device '%s' on '%s'>" %
+                    (self.device.name.strip(),
+                     self.device.platform.name.strip()))
 
     test_plat_and_dev = get_test_platforms_and_devices()
 
@@ -286,7 +274,7 @@ def pytest_generate_tests_for_pyopencl(metafunc):
 
                 metafunc.addcall(funcargs=arg_dict.copy(),
                         id=", ".join("%s=%s" % (arg, value)
-                                for arg, value in arg_dict.iteritems()))
+                                for arg, value in six.iteritems(arg_dict)))
 
     elif "platform" in metafunc.funcargnames:
         for platform, plat_devs in test_plat_and_dev:
@@ -352,8 +340,8 @@ def parse_c_arg(c_arg, with_offset=False):
     c_arg = c_arg.replace("__global", "")
 
     if with_offset:
-        vec_arg_factory = lambda dtype, name: \
-                VectorArg(dtype, name, with_offset=True)
+        def vec_arg_factory(dtype, name):
+            return VectorArg(dtype, name, with_offset=True)
     else:
         vec_arg_factory = VectorArg
 
@@ -470,11 +458,11 @@ class _CDeclList:
         if dtype in self.declared_dtypes:
             return
 
-        from pyopencl.array import vec
-        if dtype in vec.type_to_scalar_and_count:
+        import pyopencl.cltypes
+        if dtype in pyopencl.cltypes.vec_type_to_scalar_and_count:
             return
 
-        for name, field_data in dtype.fields.iteritems():
+        for name, field_data in sorted(six.iteritems(dtype.fields)):
             field_dtype, offset = field_data[:2]
             self.add_dtype(field_dtype)
 
@@ -552,11 +540,12 @@ def match_dtype_to_c_struct(device, name, dtype, context=None):
     function, not the original one.
     """
 
-    fields = sorted(dtype.fields.iteritems(),
-            key=lambda (name, (dtype, offset)): offset)
+    fields = sorted(six.iteritems(dtype.fields),
+            key=lambda name_dtype_offset: name_dtype_offset[1][1])
 
     c_fields = []
-    for field_name, (field_dtype, offset) in fields:
+    for field_name, dtype_and_offset in fields:
+        field_dtype, offset = dtype_and_offset[:2]
         c_fields.append("  %s %s;" % (dtype_to_ctype(field_dtype), field_name))
 
     c_decl = "typedef struct {\n%s\n} %s;\n\n" % (
@@ -564,25 +553,26 @@ def match_dtype_to_c_struct(device, name, dtype, context=None):
             name)
 
     cdl = _CDeclList(device)
-    for field_name, (field_dtype, offset) in fields:
+    for field_name, dtype_and_offset in fields:
+        field_dtype, offset = dtype_and_offset[:2]
         cdl.add_dtype(field_dtype)
 
     pre_decls = cdl.get_declarations()
 
     offset_code = "\n".join(
             "result[%d] = pycl_offsetof(%s, %s);" % (i+1, name, field_name)
-            for i, (field_name, (field_dtype, offset)) in enumerate(fields))
+            for i, (field_name, _) in enumerate(fields))
 
     src = r"""
         #define pycl_offsetof(st, m) \
-                 ((size_t) ((__local char *) &(dummy.m) \
+                 ((uint) ((__local char *) &(dummy.m) \
                  - (__local char *)&dummy ))
 
         %(pre_decls)s
 
         %(my_decl)s
 
-        __kernel void get_size_and_offsets(__global size_t *result)
+        __kernel void get_size_and_offsets(__global uint *result)
         {
             result[0] = sizeof(%(my_type)s);
             __local %(my_type)s dummy;
@@ -603,7 +593,7 @@ def match_dtype_to_c_struct(device, name, dtype, context=None):
     knl = prg.build(devices=[device]).get_size_and_offsets
 
     import pyopencl.array  # noqa
-    result_buf = cl.array.empty(queue, 1+len(fields), np.uintp)
+    result_buf = cl.array.empty(queue, 1+len(fields), np.uint32)
     knl(queue, (1,), (1,), result_buf.data)
     queue.finish()
     size_and_offsets = result_buf.get()
@@ -617,11 +607,15 @@ def match_dtype_to_c_struct(device, name, dtype, context=None):
 
         if dtype.itemsize == size:
             # If sizes match, use numpy's idea of the offsets.
-            offsets = [offset
-                    for field_name, (field_dtype, offset) in fields]
+            offsets = [dtype_and_offset[1]
+                    for field_name, dtype_and_offset in fields]
         else:
             raise RuntimeError(
-                    "cannot discover struct layout on '%s'" % device)
+                    "OpenCL compiler reported offsetof() past sizeof() "
+                    "for struct layout on '%s'. "
+                    "This makes no sense, and it's usually indicates a "
+                    "compiler bug. "
+                    "Refusing to discover struct layout." % device)
 
     result_buf.data.release()
     del knl
@@ -665,13 +659,21 @@ def match_dtype_to_c_struct(device, name, dtype, context=None):
 
 @memoize
 def dtype_to_c_struct(device, dtype):
+    if dtype.fields is None:
+        return ""
+
+    import pyopencl.cltypes
+    if dtype in pyopencl.cltypes.vec_type_to_scalar_and_count:
+        # Vector types are built-in. Don't try to redeclare those.
+        return ""
+
     matched_dtype, c_decl = match_dtype_to_c_struct(
             device, dtype_to_ctype(dtype), dtype)
 
     def dtypes_match():
         result = len(dtype.fields) == len(matched_dtype.fields)
 
-        for name, val in dtype.fields.iteritems():
+        for name, val in six.iteritems(dtype.fields):
             result = result and matched_dtype.fields[name] == val
 
         return result
@@ -812,8 +814,8 @@ class _TemplateRenderer(object):
                 all_args.extend(arg_list)
 
         if with_offset:
-            vec_arg_factory = lambda typename, name: \
-                    _VectorArgPlaceholder(typename, name, with_offset=True)
+            def vec_arg_factory(typename, name):
+                    return _VectorArgPlaceholder(typename, name, with_offset=True)
         else:
             vec_arg_factory = _VectorArgPlaceholder
 
@@ -848,12 +850,12 @@ class _TemplateRenderer(object):
         if arguments is not None:
             cdl.visit_arguments(arguments)
 
-        for tv in self.type_aliases.itervalues():
+        for _, tv in sorted(six.iteritems(self.type_aliases)):
             cdl.add_dtype(tv)
 
         type_alias_decls = [
                 "typedef %s %s;" % (dtype_to_ctype(val), name)
-                for name, val in self.type_aliases.iteritems()
+                for name, val in sorted(six.iteritems(self.type_aliases))
                 ]
 
         return cdl.get_declarations() + "\n" + "\n".join(type_alias_decls)
@@ -899,7 +901,7 @@ class KernelTemplateBase(object):
     def build(self, context, *args, **kwargs):
         """Provide caching for an :meth:`build_inner`."""
 
-        cache_key = (context, args, tuple(sorted(kwargs.iteritems())))
+        cache_key = (context, args, tuple(sorted(six.iteritems(kwargs))))
         try:
             return self.build_cache[cache_key]
         except KeyError:
@@ -950,4 +952,27 @@ def array_module(a):
 
 # }}}
 
+
+def is_spirv(s):
+    spirv_magic = b"\x07\x23\x02\x03"
+    return (
+            isinstance(s, six.binary_type)
+            and (
+                s[:4] == spirv_magic
+                or s[:4] == spirv_magic[::-1]))
+
+
+# {{{ numpy key types builder
+
+class _NumpyTypesKeyBuilder(KeyBuilderBase):
+    def update_for_type(self, key_hash, key):
+        if issubclass(key, np.generic):
+            self.update_for_str(key_hash, key.__name__)
+            return
+
+        raise TypeError("unsupported type for persistent hash keying: %s"
+                % type(key))
+
+# }}}
+
 # vim: foldmethod=marker
